/*
 * ISDN accounting for isdn4linux. 
 *
 * Copyright 1996 by Stefan Luethje (luethje@konstanz.netsurf.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/****************************************************************************/

#define  _SERVER_C_

/****************************************************************************/

#include "isdnlog.h"

/****************************************************************************/

char **Old_Prints = NULL;
char **Old_Info   = NULL;
char **Cur_Info   = NULL;
int Cur_Info_Size = 0;

/****************************************************************************/

int connecting(int port);
void read_loop(int listen_socket);
int start_server(int port);
int new_client(int sock);
int Compare(char *User, char *Host);
int disconnect_client(int sock);
int User_Get_Message(int sock, char* mymsn, int Flag);
int del_Cur_Info(int channel);
int add_Cur_Info(int channel, char *String);
int append_Old_Info(char *String);
char *Get_Adress(char *Number);
int Write_Caller_Message(int sock, CALL *info);
char *Build_Call_Info(CALL *call, int chan);
int save_messages(char *String, int channel, int stat);
int save_prints(char *String);
int broadcast_from_server(char* Num, int Flag, int Message, char *String);
int print_from_server(char *String);
int append_Old_Prints(char *String);
int String_For_Output(char *String);

/****************************************************************************/

PUBLIC void Exit(int RetCode);
PUBLIC void exit_on_signal(int Sign);

/****************************************************************************/

extern time_t  cur_time;
extern double  einheit;

/****************************************************************************/

int connecting(int port)
{
	int sock;


	if ((sock = client_connect("localhost",port)) < 0)
	{
		if (sock == NO_SOCKET)
		{
			print_msg(PRT_NORMAL,"Can not open socket %d: %s\n",sock,strerror(errno));
		}
		else
		if (sock == NO_BIND)
		{
			print_msg(PRT_NORMAL,"Can not bind socket: %s\n",strerror(errno));
		}

		return -1;
	}

	return sock;
}

/****************************************************************************/

int listening(int port)
{
	int sock;
	struct servent *sp = NULL;


	if ((sock = server_connect(&sp,port)) < 0)
	{
		if (sock == NO_SOCKET)
		{
			print_msg(PRT_NORMAL,"Can not open socket %d: %s\n",sock,strerror(errno));
		}
		else
		if (sock == NO_BIND)
		{
			print_msg(PRT_NORMAL,"Can not bind socket: %s\n",strerror(errno));
		}
		else
		if (sock == NO_LISTEN)
		{
			print_msg(PRT_NORMAL,"Can not start listening: %s\n",strerror(errno));
		}

		return -1;
	}

	if (sp)
	{
		print_msg(PRT_DEBUG_CS,"Got service %s on port %d\n",
			sp->s_name,ntohs(sp->s_port));
	}
	else
	{
		print_msg(PRT_DEBUG_CS,"Got port %d\n", port);
	}

	return sock;
}
/****************************************************************************/

char *Get_Hostname(struct sockaddr *Addr)
{
	char *RetCode = NULL;
	char *Ptr;
	struct hostent *hp = NULL;
	struct in_addr *In = &((struct sockaddr_in*) Addr)->sin_addr;


	if ((hp = gethostbyaddr((char*) In,sizeof(long int),AF_INET)) != NULL)
	{
		RetCode = (char*) calloc(strlen(hp->h_name),sizeof(char));
		strcpy(RetCode,hp->h_name);
	}
	else
	if ((Ptr = inet_ntoa(*In)) != NULL)
	{
		RetCode = (char*) calloc(strlen(Ptr),sizeof(char));
		strcpy(RetCode,Ptr);
	}
	else
	{
		RetCode = (char*) calloc(strlen("Unkown"),sizeof(char));
		strcpy(RetCode,"Unkown");
	}

	return RetCode;
}

/****************************************************************************/

int new_client(int sock)
{
	int Accept;
	int Cnt;
	static CALL *Info = NULL;


	sockets[sock].f_username = (char*) calloc(strlen(sockets[sock].msgbuf.buf)+1,
		sizeof(char));
	strcpy(sockets[sock].f_username,sockets[sock].msgbuf.buf);

	Accept = Compare(sockets[sock].f_username,sockets[sock].f_hostname);

	sockets[sock].waitstatus = WF_NOTHING;

	if (Accept == -1)
	{
		sockets[sock].msg = MSG_ANN_REJ;
		sockets[sock].msgbuf.used = 0;

		Write(&(sockets[sock]));

		disconnect_client(sock);
		return 0;
	}
	else
	{
		sockets[sock].msg = MSG_ANN_ACC;
		msgcpy(&(sockets[sock]),itos(Accept,_MSG_2B),_MSG_2B+1);

		if (Write(&(sockets[sock])) < 1)
		{
		 	disconnect_client(sock);
			return 0;
		}
	}

	if (Cur_Info != NULL)
		for (Cnt = 0; Cnt < Cur_Info_Size; Cnt++)
			if (Cur_Info[Cnt] != NULL &&
			    Set_Info_Struct(&Info,Cur_Info[Cnt]) >= 0 &&
			    User_Get_Message(sock,Info->num[1],0))
			{
				sockets[sock].msg = MSG_CALL_INFO;
				msgcpy(&(sockets[sock]),Cur_Info[Cnt],strlen(Cur_Info[Cnt])+1);

				if (Write(&(sockets[sock])) < 1)
				{
				 	disconnect_client(sock);
				 	return 0;
				}
			}

	if (Old_Info != NULL)
		for (Cnt = 0; Cnt < MAX_CALLS_IN_QUEUE; Cnt++)
			if (Old_Info[Cnt] != NULL &&
			    Set_Info_Struct(&Info,Old_Info[Cnt]) >= 0 &&
			    User_Get_Message(sock,Info->num[1],0))
			{
				sockets[sock].msg = MSG_CALL_INFO;
				msgcpy(&(sockets[sock]),Old_Info[Cnt],strlen(Old_Info[Cnt])+1);

				if (Write(&(sockets[sock])) < 1)
				{
				 	disconnect_client(sock);
				 	return 0;
				}
			}

	if (Old_Prints != NULL && User_Get_Message(sock,0,ALLOW_SRV_MSG))
		for (Cnt = 0; Cnt < MAX_PRINTS_IN_QUEUE; Cnt++)
			if (Old_Prints[Cnt] != NULL)
			{
				sockets[sock].msg = MSG_SERVER;
				msgcpy(&(sockets[sock]),Old_Prints[Cnt],strlen(Old_Prints[Cnt])+1);

				if (Write(&(sockets[sock])) < 1)
				{
				 	disconnect_client(sock);
				 	return 0;
				}
			}
			else
				break;

	print_msg(PRT_DEBUG_CS,"connection from client %s@%s %s\n",
		sockets[sock].f_username,
		sockets[sock].f_hostname,
		Accept != -1?"accepted":"rejected");

	return 0;
}

/****************************************************************************/

int Compare(char *User, char *Host)
{
	/* Hier muss der User und der Host ueberprueft werden */
	/* Rejected: return -1;                               */
	return T_ISDN4LINUX | T_SERVER_MSG;
}

/****************************************************************************/

int User_Get_Message(int sock, char* mymsn, int Flag)
{
	/* Hier muss ueberprueft werden, ob der User die MSG vom Server erhalten
	   darf. */

	return 1;
}

/****************************************************************************/

int disconnect_client(int sock)
{
	char User[SHORT_STRING_SIZE];
	char Host[SHORT_STRING_SIZE];

	strcpy(User,sockets[sock].f_username);
	strcpy(Host,sockets[sock].f_hostname);

	close(sockets[sock].descriptor);

	if (del_socket(&sockets,sock))
		Exit(16);

	print_msg(PRT_DEBUG_CS,"disconnect from client %s@%s!\n",User,Host);
	print_msg(PRT_DEBUG_CS,"I/O error: %s!\n",strerror(errno));

	return 0;
}

/****************************************************************************/

int eval_message (int sock)
{
	long int Value;
	int      Msg;

	do
	{
		if ((Value = Read(&(sockets[sock]))) < 0)
		{
			disconnect_client(sock);
			break;
		}
		else
		{
			Msg = sockets[sock].msg;
			switch(sockets[sock].waitstatus)
			{
				case WF_ACC:
					if (Msg == MSG_ANNOUNCE)
					{
						new_client(sock);
					}
					else
					{
						/* Meldung: Unbekannter Message-Typ Msg */
							print_msg(PRT_ERR,"Unknown Message %d: %s\n",Msg,sockets[sock].msgbuf.buf);
					}
					break;

				default:
					switch (Msg)
					{
						default:
							/* Meldung: Unbekannter Message-Typ Msg */
							print_msg(PRT_ERR,"Unknown Message %d: %s\n",Msg,sockets[sock].msgbuf.buf);
							break;
					}
					break;
			}
		}
	} while (sockets[sock].status == NEXT_MSG);


	return 0;
}

/****************************************************************************/

int broadcast_from_server(char* Num, int Flag, int Message, char *String)
{
	int Cnt;

	for(Cnt = FIRST_DESCR; Cnt < socket_size(sockets); Cnt++)
	{
		if (sockets[Cnt].fp == NULL && User_Get_Message(Cnt,Num,Flag))
		{
			sockets[Cnt].msg = Message;
			msgcpy(&(sockets[Cnt]),String,strlen(String)+1);

			if (Write(&(sockets[Cnt])) < 1)
 				disconnect_client(Cnt);
 		}
	}

	return  0;
}

/****************************************************************************/

int print_from_server(char *String)
{
	int RetCode;
	char NewString[LONG_STRING_SIZE];
	time_t t = time(NULL);
	struct tm *tm_time = localtime(&t);


	tm_time->tm_isdst = 0;

	strftime(NewString,LONG_STRING_SIZE,"%b %d %X ",tm_time);
	strcat(NewString,String);

	if ((RetCode = String_For_Output(NewString)) < 0)
		return RetCode;

	if ((RetCode = append_Old_Prints(NewString)) < 0)
		return RetCode;

	return broadcast_from_server(NULL,T_SERVER_MSG,MSG_SERVER,NewString);
}

/****************************************************************************/

int message_from_server(CALL *call, int chan)
{
	int Cnt;
	int RetCode;
	char *String = Build_Call_Info(call,chan);

	if ((RetCode = String_For_Output(String)) < 0)
		return RetCode;

	if ((RetCode = save_messages(String,chan,call->stat)) < 0)
		return RetCode;

	for(Cnt = FIRST_DESCR; Cnt < socket_size(sockets); Cnt++)
	{
		if (sockets[Cnt].fp == NULL && User_Get_Message(Cnt,call->num[ME],0))
		{
			sockets[Cnt].msg = MSG_CALL_INFO;
			msgcpy(&(sockets[Cnt]),String,strlen(String)+1);

			if (Write(&(sockets[Cnt])) < 1)
 				disconnect_client(Cnt);
 			else
 			if (Write_Caller_Message(Cnt,call) < 0)
 				disconnect_client(Cnt);
 		}
	}

	return 0;
}

/****************************************************************************/

char *Build_Call_Info(CALL *call, int chan)
{
	static char RetCode[LONG_STRING_SIZE];

/* Anmerkungen dazu:

     1. ``connect'' kann auch (time_t)0 sein!
     	Dann besteht noch gar keine Verbindung - z.b. weil es gerade
        klingelt, man aber noch nicht abgenommen hat!

     2. Die ``Dauer'' zu uebertragen halte ich fuer unwichtig - bitte
     	korrigiert mich, wenn ich da falsch liege!
        Die Dauer ist doch ``(jetzt) - connect'' ?

     3. Die Einheiten sind:

     	  0         : kostet nix (0130 Gespraech, oder man wird angerufen)
          -1 .. -x  : Gebuehrenimpuls
          1 .. x    : AOCE (Gebuehreninfo am Ende)
*/

      sprintf(RetCode, "%d|%d|%d|%s|%s|%s|%s|%s|%s|%d|%d|%d|%s|%s|%ld|%ld|%g|%g|%s\n",
        chan,                                                                                    /* Kennung: 0 = erste, 1 = zweite laufende Verbindung */
        call->stat,			     	 						       /* Satzart fuer _diese_  Meldung */
        call->dialin,	     	 						       /* 1 = ein Anruf, 0 = man waehlt raus */
        call->num[ME],            	 						       /* Eigene MSN */
        call->num[OTHER],         	 						       /* Telefonnummer des Anrufers/Angerufenen (wie von der VSt uebermittelt) */
        call->vorwahl[OTHER],     	 						       /* Vorwahl des Anrufers/Angerufenen */
        call->rufnummer[OTHER],   	 						       /* Rufnummer des Anrufers/Angerufenen */
        call->alias[OTHER], 	     	 						       /* Alias des Anrufers/Angerufenen */
        call->area[OTHER], 	     	 						       /* Ortsnetz des Anrufers/Angerufenen */
        (int)call->connect,       	 						       /* Beginn (als (time_t) ) */
        (int)(cur_time - call->connect),				       /* aktuelle Dauer - in Sekunden seit CONNECT */
        call->aoce,		     	                                                       /* Einheiten (negativ: laufende Impulse, positiv: AOCE) */
        double2str(abs(call->aoce) * (currency_factor ? currency_factor : einheit), 6, 2, DEB), /* kostet gerade */
        currency_factor ? currency : "DM", 	       		       	 	      	       /* Waehrung */
        call->ibytes,	     	 						       /* Frank's ibytes */
        call->obytes,	     	 						       /* Frank's obytes */
        call->ibps,									       /* Aktueller Durchsatz INPUT: Bytes/Sekunde */
        call->obps,									       /* Aktueller Durchsatz OUTPUT: Bytes/Sekunde */
		    call->msg);			     	 						       /* Status */

	return RetCode;
}

/****************************************************************************/

int Write_Caller_Message(int sock, CALL *Info)
{
#if 0
	char *Address = NULL;

	/* Dirty Hack fuer emulation von neuen Anrufern */
	if (Info->stat == CS_DISCONN || Info->stat == CS_ERROR)
 	{
 		if ((Address = Get_Adress(Info->num[0])) == NULL)
		{
			sockets[sock].msg = MSG_WHO_IS;
			msgcpy(&(sockets[sock]),Info->num[0],strlen(Info->num[0])+1);
/* Nur fuer Test */
msgcpy(&(sockets[sock]),"984720",7);

			if (Write(&(sockets[sock])) < 1)
			{
 				disconnect_client(sock);
 				return -1;
 			}
 		}
 		else
 		{
			sockets[sock].msg = MSG_CALLER;
			msgcpy(&(sockets[sock]),Address,strlen(Address)+1);

			if (Write(&(sockets[sock])) < 1)
			{
 				disconnect_client(sock);
 				return -1;
 			}
 		}
 	}
#endif

	return 0;
}

/****************************************************************************/

char *Get_Adress(char *Number)
{
	/* Hier sollte die Adresse zur uebergebenen Telefonummer geliefert werden */
#if 0
	return "Burmester:Fred:2::Lohe 4:D:23869:Fischbek:2:04532/123:04532/124:0:0::Elmenhorster Str. 16:D:20000:Hamburg:0:0:1:fred@wo.auch.immer:08.01.68";
#else
	return NULL;
#endif
}

/****************************************************************************/

int save_messages(char *String, int channel, int stat)
{
	int RetCode;


	if (stat == CS_ERROR || stat == CS_DISCONN)
	{
		if ((RetCode = del_Cur_Info(channel)) < 0)
		{
			print_msg(PRT_DEBUG_CS,"Invalid Channel No. %d\n",channel);
		}

		if ((RetCode = append_Old_Info(String)) < 0)
			return RetCode;
	}
	else
	{
		if ((RetCode = add_Cur_Info(channel,String)) < 0)
			return RetCode;
	}

	return 0;
}

/****************************************************************************/

int del_Cur_Info(int channel)
{
	if (Cur_Info == NULL || Cur_Info[channel] == NULL ||
	    channel >= Cur_Info_Size || channel < 0         )
		return -1;

	free(Cur_Info[channel]);
	Cur_Info[channel] = NULL;
	return 0;
}

/****************************************************************************/

int append_Old_Prints(char *String)
{
	int Cnt = 0;

	if (Old_Prints == NULL)
		if ((Old_Prints = (char**) calloc(MAX_PRINTS_IN_QUEUE,sizeof(char*))) == NULL)
			return NO_MEMORY;


	while(Cnt < MAX_PRINTS_IN_QUEUE && Old_Prints[Cnt] != NULL)
		Cnt++;

	if (Cnt == MAX_PRINTS_IN_QUEUE)
	{
		Cnt = 1;

		if (Old_Prints[0] != NULL)
			free(Old_Prints[0]);

		while(Cnt < MAX_PRINTS_IN_QUEUE)
		{
			Old_Prints[Cnt-1] = Old_Prints[Cnt];
			Cnt++;
		}

		Cnt = MAX_PRINTS_IN_QUEUE - 1;
	}

	if ((Old_Prints[Cnt] = (char*) calloc(strlen(String)+1,sizeof(char))) == NULL)
		return NO_MEMORY;
	else
		strcpy(Old_Prints[Cnt],String);

	return 0;
}

/****************************************************************************/

int append_Old_Info(char *String)
{
	int Cnt = 0;


	if (Old_Info == NULL)
		if ((Old_Info = (char**) calloc(MAX_CALLS_IN_QUEUE,sizeof(char*))) == NULL)
			return NO_MEMORY;


	while(Cnt < MAX_CALLS_IN_QUEUE && Old_Info[Cnt] != NULL)
		Cnt++;

	if (Cnt == MAX_CALLS_IN_QUEUE)
	{
		Cnt = 1;
		free(Old_Info[0]);

		while(Cnt < MAX_CALLS_IN_QUEUE)
		{
			Old_Info[Cnt-1] = Old_Info[Cnt];
			Cnt++;
		}

		Cnt = MAX_CALLS_IN_QUEUE - 1;
	}

	if ((Old_Info[Cnt] = (char*) calloc(strlen(String)+1,sizeof(char))) == NULL)
		return NO_MEMORY;
	else
		strcpy(Old_Info[Cnt],String);

	return 0;
}

/****************************************************************************/

int add_Cur_Info(int channel, char *String)
{
	char **Ptr;


	if (channel < 0)
		return 0;

	if (Cur_Info == NULL)
		if ((Cur_Info = (char**) calloc(channel + 1,sizeof(char*))) == NULL)
			return NO_MEMORY;
		else
			Cur_Info_Size = channel + 1;
	else
	if (channel >= Cur_Info_Size)
	{
		if ((Ptr = (char**) calloc(channel + 1,sizeof(char*))) == NULL)
			return NO_MEMORY;

		memcpy(Ptr,Cur_Info,Cur_Info_Size*sizeof(char*));
		Cur_Info_Size = channel + 1;
		free(Cur_Info);
		Cur_Info = Ptr;
	}

	if (Cur_Info[channel] != NULL)
		free(Cur_Info[channel]);

	if ((Cur_Info[channel] = (char*) calloc(strlen(String)+1,sizeof(char))) == NULL)
		return NO_MEMORY;
	else
		strcpy(Cur_Info[channel],String);

	return 0;
}

/****************************************************************************/

int String_For_Output(char* String)
{
	char *Ptr;
	char *CurPtr;
	char  NewString[LONG_STRING_SIZE];

	strcpy(NewString,String);
	CurPtr = NewString;
	String[0] = '\0';

	while ((Ptr = strtok(CurPtr,"\n")) != NULL)
	{
		CurPtr = NULL;
		strcat(String,Ptr);
	}

	return 0;
}

/****************************************************************************/

int change_channel(int old_chan, int new_chan)
{
	int RetCode = 0;
	int Cnt;
	char tmp_string[SHORT_STRING_SIZE];
	char *String;
	static CALL *Info = NULL;


	if (old_chan < 0 || old_chan >= Cur_Info_Size ||
	    new_chan < 0 || Cur_Info[old_chan] == NULL  )
		return -1;

	if (old_chan == new_chan)
		return 0;

 	if ((RetCode = Set_Info_Struct(&Info,Cur_Info[old_chan])) < 0)
		return RetCode;

 	if ((String = Build_Call_Info(Info, new_chan)) == NULL)
		return -1;

	if ((RetCode = del_Cur_Info(old_chan)) != 0)
		return RetCode;

	if ((RetCode = add_Cur_Info(new_chan,String)) != 0)
		return RetCode;

	for(Cnt = FIRST_DESCR; Cnt < socket_size(sockets); Cnt++)
	{
	  if (User_Get_Message(Cnt,Info->num[1],0))
		{
			sockets[Cnt].msg = MSG_CHANGE_CHAN;
			sprintf(tmp_string,"%d%c%d",old_chan,TRENN,new_chan);
			msgcpy(&(sockets[Cnt]),tmp_string,strlen(tmp_string)+1);

			if (Write(&(sockets[Cnt])) < 1)
			 	disconnect_client(Cnt);
		}
	}

	return 0;
}

/****************************************************************************/

