/* $Id: start_prog.c,v 2.21 1996/03/13 11:58:46 root Exp $
 *
 * ISDN accounting for isdn4linux.
 *
 * Copyright 1996 by Michael 'Ghandi' Herold,
 *                   Stefan Luethje (luethje@konstanz.netsurf.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: start_prog.c,v $
 * Revision 2.21  1996/03/13  11:58:46  akool
 */


#define _START_PROG_C_
#include "isdnlog.h"


/** Prototypes ***********************************************************/

static void		KillCommand(pid_t);
static int		GetArgs(char *, char *[], int);
static int		ProgramExist(char *);

/*************************************************************************
 ** Ring(-) - Externes Kommando ausfuehren oder loeschen. Bei einem Fehl-	    **
 **			  er wird -1 zurueckgegeben; ansonsten die PID des neuen    **
 **			  Prozesses. Wenn das Kommando mit system() aufgerufen wur- **
 **			  de (Async), wird -1 zurueckgegeben, da die PID nicht ge-  **
 **			  sichert werden muss.					    **
 *************************************************************************
 ** Die  : <= 0   Kommando ausfuehren.						    **
 **        >  0   PID killen.							    **
 ** Async: == 1   Kommando mit system() starten.				    **
 **        == 0   Kommando mit eigenem Prozess starten.				    **
 *************************************************************************/

int Ring(char *Cmd, int Die, int Async)
{
	char  Command[SHORT_STRING_SIZE + 1];
	char  String[LONG_STRING_SIZE];
	int   filedes[2];
	char *Args[64];
	pid_t pid;
	FILE *fp;

	if (Die <= 0)
	{
		sprintf(Command,"%.*s",SHORT_STRING_SIZE-6,Cmd);

		if (!ProgramExist(Command)) return(-1);

			/* Kommando soll gestartet werden - 'Async' gibt an, ob dieses	*/
			/* mit system() oder mit einem eigenen Proze aufgerufen wer-	*/
			/* den soll.																	*/

		pipe(filedes);
		switch (pid = fork())
		{
			case -1: print_msg(PRT_ERR, "%s\n", "Can't start fork()!");
			         return 0;
			         break;
			case  0: dup2(filedes[1],STDOUT_FILENO);
			         dup2(filedes[1],STDERR_FILENO);
			         GetArgs(Command,Args,64);
			         execvp(Args[0], Args);
			         print_msg(PRT_ERR, "Can't start \"%s\" with execvp().\n", Command);
			         Alarm();
			         return 0;
			         break;
		}

		fp = fdopen(filedes[0],"r");
		close(filedes[1]);

		print_msg(PRT_INFO,"Program \"%s\" started %ssynchronous.\n", Command, (Async?"a":""));

		if (Async == 1)
		{
			while(fgets(String,LONG_STRING_SIZE,fp) != NULL)
				print_msg(PRT_PROG_OUT,"%s\n",String);

			waitpid(pid,NULL,0);
			fclose(fp);
			return(0);
		}
		else
		{
			int sock;

			if (add_socket(&sockets,filedes[0]))
				return NO_MEMORY;

			sock = socket_size(sockets)-1;
			sockets[sock].fp  = fp;
			sockets[sock].pid = pid;

			return sock;
		}
	}
	else KillCommand(Die);

	return(0);
}

/*************************************************************************
 ** GetArgs(-) - Zerlegt eine Kommandozeile in einzelne Argumente.		**
 *************************************************************************/

static int GetArgs(char *Line, char *Args[], int MaxArgs)
{
	char	*Arg	= NULL;
	char	*Use	= Line;
	int	 i		= 0;

	while ((Arg = strtok(Use, " \t")))
	{
		Use = NULL;

		if (i < MaxArgs) Args[i++] = Arg;
	}

	Args[i] = NULL;

	return(i);
}

/*************************************************************************
 ** KillCommand(-) - Beendet ein Programm anhand seiner PID.				**
 *************************************************************************/

static void KillCommand(int sock)
{
		/* Kein Erbarmen - Alles was uns zwischen die Finger kommt wird	*/
		/* gelscht :-)																	*/

	if (sock > 0)
	{
		kill(sockets[sock].pid, SIGTERM);
		kill(sockets[sock].pid, SIGKILL);

    /* ACHTUNG: Die naechste Zeile kann schwierigkeiten machen!!!
       Der Prozess kann eventuell hier haengen bleiben. Dann Zeile
       auskommentieren. Ich weiss nicht, ob kill den Prozess auch
       sauber beendet. Damit keine Zombies rumfahren vorsichtshaber
       der waitpid.
       Alternativ: waitpid(sockets[sock].pid,NULL,WNOHANG) */

		waitpid(sockets[sock].pid,NULL,0);
		fclose(sockets[sock].fp);
		del_socket(&sockets,sock);
	}
}

/*************************************************************************
 ** Alarm(-) - Gibt ein Alarmsignal ber den internen Lautsprecher aus.	**
 *************************************************************************/

void Alarm(void)
{
#ifdef ALARM

	int FD;
	int i;

	if ((FD = open("/dev/console", O_WRONLY)) == -1) FD = 0;

	for (i = 0; i < 30; i++)
	{
		ioctl(FD, KIOCSOUND, (3000 - (i * 10)));

		usleep((1 * 1000));
	}

	ioctl(FD, KIOCSOUND, 0);

#endif
}

/*************************************************************************
 ** CheckTime(-) - Prft ob die Zeitangabe in den lokalen Zeitrahmen		**
 **					 fllt. Rckgabe ist TRUE/FALSE.								**
 *************************************************************************/

int CheckTime(char *Line)
{
	char	 Temp[SHORT_STRING_SIZE + 1];
	char	 Time[24];
	char	*Use;
	char	*Arg;
	char	*Minus;
	int	 i;
	int	 r;
	int	 Beg;
	int	 End;
	time_t Local;

	strncpy(Temp, Line, SHORT_STRING_SIZE);

	for (i = 0; i < 24; i++) Time[i] = 0;

	Use = Temp;

		/* Zeile in die einzelnen Komponenten trennen, die durch ',',	*/
		/* ' ' oder ';' voneinander getrennt sein knnen.					*/

	while ((Arg = strtok(Use, ", ;")))
	{
		Use = NULL;

		if ((Minus = strchr(Arg, '-')))
		{
			*Minus++ = 0;

			Beg = atoi(Arg);
			End = atoi(Minus);
		}
		else
		{
			Beg = atoi(Arg);
			End = Beg;
		}

		if ((Beg < 0) || (End < 0) || (Beg > 23) || (End > 23) || (End < Beg))
		{
			print_msg(PRT_WARN, "Time range not correct (%d-%d).\n", Beg, End);
		}
		else
		{
			for (r = Beg; r <= End; r++) Time[r] = 1;
		}
	}

		/* Lokale Zeit errechnen und mit den Stunden im Zeitarray ver-	*/
		/* gleichen.																	*/

	Local = time(NULL);

	if ((Arg = ctime(&Local)))
	{
		Beg = atoi(&Arg[11]);

		if ((Beg >= 0) && (Beg <= 23))
		{
			if (Time[Beg] == 1) return(1);
		}
		else print_msg(PRT_ERR, "Bad hour in local time.\n");
	}
	else print_msg(PRT_ERR, "Can't get local time.\n");

	return(0);
}

/*************************************************************************
 ** ProgramExist(-) - Prft ob ein Programm existiert.						**
 *************************************************************************/

static int ProgramExist(char *Cmd)
{
	struct stat Status;

	char	Name[SHORT_STRING_SIZE + 1];
	char *Temp;

	if (Cmd)
	{
		strncpy(Name, Cmd, SHORT_STRING_SIZE);

		if ((Temp = strchr(Name, ' '))) *Temp = 0;

		if (stat(Name, &Status) != 0)
		{
			print_msg(PRT_ERR, "Command \"%s\" doesn't exist.\n", Name);

			Alarm();
		}
		else return(1);
	}

	return(0);
}

/****************************************************************************/

int Print_Cmd_Output( int sock )
{
	char String[LONG_STRING_SIZE] = "";


	if (feof(sockets[sock].fp))
		return del_socket(&sockets,sock);

	fgets(String,LONG_STRING_SIZE,sockets[sock].fp);

	print_msg(PRT_PROG_OUT,"%s\n",String);

	return 0;
}

/****************************************************************************/

